// PV112 2017, lesson 2

#include <iostream>
#include "PV112.h"

#define _USE_MATH_DEFINES
#include <cmath>

// Include GLEW, use static library
#define GLEW_STATIC
#include "GL/glew.h"
#if defined(_WIN32)
#define NOMINMAX // Make Windows.h not define 'min' and 'max' macros
#include <GL/wglew.h> // Include on Windows
#pragma comment(lib, "glew32s.lib")
#else
#include <GL/glxew.h> // Include on Linux and Mac
#define APIENTRY
#endif

#include <GL/freeglut.h>

#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>

using namespace std;
using namespace PV112;

// Current window size
int win_width = 480;
int win_height = 480;

GLuint program;

GLint color_loc;
GLint PVM_matrix_loc;
GLint normal_matrix_loc;

// Simple geometries that we will use in this lecture
PV112Geometry my_cube;
PV112Geometry my_sphere;
PV112Geometry my_teapot;

// Current time of the application in seconds, for animations
float app_time_s = 0.0f;

// Called when the user presses a key
void key_pressed(unsigned char key, int mouseX, int mouseY)
{
    switch (key)
    {
    case 27: // Escape
     exit(0);
     break;
    case 'l':
     glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
     glutPostRedisplay();
     break;
    case 'f':
     glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
     glutPostRedisplay();
     break;
    case 't':
     glutFullScreenToggle();
     break;
    }
}

// Initializes OpenGL stuff
void init()
{
    // Color to clear the frame buffer with
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    // Value to clear the depth buffer with
    glClearDepth(1.0);

    // Actually make use of the depth buffer
    glEnable(GL_DEPTH_TEST);

    // Create the shader program
    program = CreateAndLinkProgram("vertex.glsl", "fragment.glsl");
    if (0 == program)
        WaitForEnterAndExit();

    // Get attribute and uniform locations
    int position_loc = glGetAttribLocation(program, "position");
    int normal_loc = glGetAttribLocation(program, "normal");
    color_loc = glGetUniformLocation(program, "color");
    PVM_matrix_loc = glGetUniformLocation(program, "PVM_matrix");
    normal_matrix_loc = glGetUniformLocation(program, "normal_matrix");

    my_cube = CreateCube(position_loc, normal_loc);
    my_sphere = CreateSphere(position_loc, normal_loc);
    my_teapot = CreateTeapot(position_loc, normal_loc);
}

// Called when the window needs to be rerendered
void render()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Matrices we use to transform coordinates
    glm::mat4 projection_matrix, view_matrix, model_matrix, PVM_matrix;

    // Needed for our simple light, we will learn this the next lesson
    glm::mat3 normal_matrix;

    glUseProgram(program);

    // View matrix basically represents a camera.
    // glm::lookAt creates the matrix so that we look from the 'eye' position
    // in the direction of the 'center' point. The 'up' vector is used to
    // rotate the camera around the sagittal axis and must be normalised
    view_matrix = glm::lookAt(glm::vec3{10.0f, 5.5f, 15.0f},// eye
                              glm::vec3{0.0f, 1.5f, -1.0f}, // center
                              glm::vec3{0.0f, 1.0f, 0.0f}); // up vector

    // Projection matrix defines how our 3D scene projects onto screen.
    // glm::perspective is used to create perspective projection (where distant
    // object seem smaller). An alternative could be the orthographic projection
    // (glm::ortho). The matrix is created in such way, that it deals with the
    // window aspect ratio and also flips the z axis (providing us with a
    // right-handed coordinate system).
    projection_matrix = glm::perspective(
            glm::radians(45.0f), // fovy (vertical field of view)
            float(win_width) / float(win_height), // window aspect ratio
            0.1f, // near clipping plane
            40.0f); // far clipping plane


    // The following code is what a solved task 10 could look like.
    // TASK 10 (optional): Using non-uniform scale on cube, create a simple
    // desk. You don't have to create new cubes, just make more draw calls with
    // different matrices. Place two spheres and a rotating teapot on top of
    // the desk. Feel free to play with the transformations; e.g. let one of
    // the spheres jump on the table and the other one change its size over
    // time. You can utilize for example the sine function.

    //------------
    //-- TABLE --

    glBindVertexArray(my_cube.VAO);

    glUniform3f(color_loc, 1.0f, 1.0f, 1.0f);

    // Model matrix defines transformations on each object, e.g. where it is
    // located in the world, how it is rotated or scaled.
    // The first parameter of the glm::translate/rotate/scale function is a
    // matrix which the transformation is applied to. If you had a matrix for
    // each of the transformations, the following sequence of functions would
    // be equivalent to "model_matrix = scale_matrix * translate_matrix;"
    // However, the transformations are applied from the right (or from the
    // bottom when using this notation). Here, we first move the cube, so that
    // center of coordinate system (i.e. [0,0,0]) is in the cube's top face.
    // Then we scale the cube non-uniformly, transforming it into a cuboid. The
    // center of its top face is still at [0,0,0].
    model_matrix = glm::mat4(1.0f);
    model_matrix = glm::scale(model_matrix, glm::vec3(0.4f, 3.0f, 0.4f));
    model_matrix = glm::translate(model_matrix, glm::vec3(0.0f, -1.0f, 0.0f));

    // PVM matrix combines all of the transformation matrices
    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));

    // Needed for our simple light, we will learn this the next lesson
    normal_matrix = glm::inverse(glm::transpose(glm::mat3(view_matrix * model_matrix)));
    glUniformMatrix3fv(normal_matrix_loc, 1, GL_FALSE, glm::value_ptr(normal_matrix));

    DrawGeometry(my_cube);

    // We draw another cuboid using the same object, we only change the model
    // matrix, so that it looks different.
    model_matrix = glm::mat4(1.0f);
    model_matrix = glm::scale(model_matrix, glm::vec3(5.0f, 0.2f, 5.0f));
    model_matrix = glm::translate(model_matrix, glm::vec3(0.0f, 1.0f, 0.0f));

    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    normal_matrix = glm::inverse(glm::transpose(glm::mat3(view_matrix * model_matrix)));
    glUniformMatrix3fv(normal_matrix_loc, 1, GL_FALSE, glm::value_ptr(normal_matrix));

    DrawGeometry(my_cube);

    //--------------
    //-- SPHERES --

    glBindVertexArray(my_sphere.VAO);
    glUniform3f(color_loc, 1.0f, 0.0f, 0.0f);

    // Two translations are used. Translation (1) is applied first and it
    // effectively changes the scale's (2) center (or origin) to the sphere's
    // lowest point. Then we scale (2), knowing that the bottom point (which is
    // [0,0,0]) doesn't move away. Then we translate (3) the sphere again, now
    // to actually move the sphere. After that, we know the bottom point has
    // coordinates of [-2.7, 0.4, 1.0]. We used two translations to make sure
    // the sphere sits precisely on the table.
    float scale_factor = 1.0f + 0.2 * sin(3.0f * app_time_s);
    model_matrix = glm::mat4(1.0f);
    model_matrix = glm::translate(model_matrix, {-2.7f, 0.4f, 1.0f}); // (3)
    model_matrix = glm::scale(model_matrix, glm::vec3(scale_factor)); // (2)
    model_matrix = glm::translate(model_matrix, {0.0f, 1.0f, 0.0f});  // (1)

    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    normal_matrix = glm::inverse(glm::transpose(glm::mat3(view_matrix * model_matrix)));
    glUniformMatrix3fv(normal_matrix_loc, 1, GL_FALSE, glm::value_ptr(normal_matrix));

    DrawGeometry(my_sphere);

    // Another sphere, we use the same vertex data, but different model matrix
    glUniform3f(color_loc, 1.0f, 0.4f, 0.0f);

    // Ideal bouncing
    float v0 = 7.0f;
    float g = 9.8f;
    float tmax = 2.0f * v0 / g;
    float t = app_time_s;
    while (t >= tmax) t -= tmax;
    float jump = 1.4f + v0 * t - 0.5f * g * t * t;

    model_matrix = glm::mat4(1.0f);
    model_matrix = glm::translate(model_matrix, {2.0f, jump, 3.2f});

    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    normal_matrix = glm::inverse(glm::transpose(glm::mat3(view_matrix * model_matrix)));
    glUniformMatrix3fv(normal_matrix_loc, 1, GL_FALSE, glm::value_ptr(normal_matrix));

    DrawGeometry(my_sphere);

    //--------------
    //-- TEAPOT --

    glBindVertexArray(my_teapot.VAO);

    glUniform3f(color_loc, 0.0f, 1.0f, 1.0f);

    model_matrix = glm::mat4();
    model_matrix = glm::translate(model_matrix, {2.3f, 0.4f, -2.7f});
    model_matrix = glm::rotate(model_matrix, app_time_s * 2, {0.0f, 1.0f, 0.0f});

    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    normal_matrix = glm::inverse(glm::transpose(glm::mat3(view_matrix * model_matrix)));
    glUniformMatrix3fv(normal_matrix_loc, 1, GL_FALSE, glm::value_ptr(normal_matrix));

    DrawGeometry(my_teapot);

    glBindVertexArray(0);
    glUseProgram(0);

    glutSwapBuffers();
}

// Called when the window changes its size
void reshape(int width, int height)
{
    win_width = width;
    win_height = height;

    // Set the area into which we render
    glViewport(0, 0, win_width, win_height);
}

// Callback function to be called when we make an error in OpenGL
void APIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id,
        GLenum severity, GLsizei length, const char* message, const void* userParam)
{
    switch (type) // Report only some errors
    {
    case GL_DEBUG_TYPE_ERROR:
    case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
    case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
     cout << message << endl; // Put the breakpoint here
     return;
    default:
     return;
    }
}

// Simple timer function for animations
void timer(int)
{
    // Add 20 ms to the application time, register the timer to be called again
    // in 20 ms, and rerender the scene.
    app_time_s += 0.020f;
    glutTimerFunc(20, timer, 0);
    glutPostRedisplay();
}

int main(int argc, char ** argv)
{
    // Initialize GLUT
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

    // Set OpenGL Context parameters
    glutInitContextVersion(3, 3); // Use OpenGL 3.3
    glutInitContextProfile(GLUT_CORE_PROFILE); // Use OpenGL core profile
    glutInitContextFlags(GLUT_DEBUG); // Use OpenGL debug context

    // Initialize window
    glutInitWindowSize(win_width, win_height);
    glutCreateWindow("PV112 - cv2");

    // Initialize GLEW
    glewExperimental = GL_TRUE;
    glewInit();

    SetDebugCallback(simple_debug_callback);

    // Initialize our OpenGL stuff
    init();

    // Register callbacks
    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(key_pressed);
    glutTimerFunc(20, timer, 0);

    // Run the main loop
    glutMainLoop();

    return 0;
}
